#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com


import os
from tempfile import NamedTemporaryFile


class BaseOptimizer(object):
    def __init__(self, context):
        self.context = context

    def should_run(self, image_extension, buffer):
        return True

    def run_optimizer(self, image_extension, buffer):
        if not self.should_run(image_extension, buffer):
            return buffer

        ifile = NamedTemporaryFile(delete=False)
        ofile = NamedTemporaryFile(delete=False)
        try:
            ifile.write(buffer)
            ifile.close()
            ofile.close()

            self.optimize(buffer, ifile.name, ofile.name)

            with open(ofile.name, 'rb') as f:  # reopen with file thats been changed with the optimizer
                return f.read()

        finally:
            os.unlink(ifile.name)
            os.unlink(ofile.name)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from thumbor.filters import BaseFilter, filter_method


class Filter(BaseFilter):

    @filter_method(BaseFilter.DecimalNumber)
    def proportion(self, value):
        source_width, source_height = self.context.request.engine.size

        new_width = source_width * value
        new_height = source_height * value

        if new_width <= 0 or new_height <= 0:
            return

        self.engine.resize(new_width, new_height)



import signal
import sys
import time

from thumbor.utils import logger
import tornado.ioloop


def signal_handler(server, config, sig, frame):
    io_loop = tornado.ioloop.IOLoop.instance()

    def stop_loop(now, deadline):
        if now < deadline and (io_loop._callbacks or io_loop._timeouts):
            logger.debug('Waiting for next tick')
            now += 1
            io_loop.add_timeout(now, stop_loop, now, deadline)
        else:
            io_loop.stop()
            logger.debug('Shutdown finally')

    def stop_server(now, deadline):
        if now < deadline:
            logger.debug('Waiting for next tick')
            now += 1
            io_loop.add_timeout(now, stop_server, now, deadline)
        else:
            server.stop()
            logger.debug('Stopped http server.')
            logger.debug('Will shutdown io in maximum %d seconds ...',
                         config.MAX_WAIT_SECONDS_BEFORE_IO_SHUTDOWN)
            now = time.time()
            stop_loop(now, now + config.MAX_WAIT_SECONDS_BEFORE_IO_SHUTDOWN)

    def shutdown():
        logger.debug('Stopping http server (i.e. stopping accepting new requests) in %d seconds ...',
                     config.MAX_WAIT_SECONDS_BEFORE_SERVER_SHUTDOWN)
        now = time.time()
        stop_server(now, now + config.MAX_WAIT_SECONDS_BEFORE_SERVER_SHUTDOWN)

    logger.warning('Caught signal: %d. Shutting down server.', sig)
    io_loop.add_callback_from_signal(shutdown)


def setup_signal_handler(server, config):

    def server_sig_handler(sig, frame):
        signal_handler(server, config, sig, frame)
        if sig == signal.SIGINT:
            sys.stdout.write('\n')
            sys.stdout.write("-- thumbor closed by user interruption --\n")

    signal.signal(signal.SIGTERM, server_sig_handler)
    signal.signal(signal.SIGINT, server_sig_handler)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import os
import subprocess
import re
import webcolors

from thumbor.optimizers import BaseOptimizer
from thumbor.utils import logger
from os.path import exists


class Optimizer(BaseOptimizer):

    def should_run(self, image_extension, buffer):
        if 'gif' in image_extension and 'gifv' in self.context.request.filters:
            if not exists(self.context.config.FFMPEG_PATH):
                logger.warn(
                    'gifv optimizer enabled but binary FFMPEG_PATH does not exist')
                return False
            return True
        return False

    def optimize(self, buffer, input_file, output_file):
        format, command_params = self.set_format()

        bg_color_hex = None
        if 'background_color' in self.context.request.filters:
            filters = self.context.request.filters.split(':')
            bg_filter = [filter for filter in filters if filter.startswith(
                'background_color')][0]
            bg_color = re.search(r'\((.*?)\)', bg_filter).group(1)
            bg_color_hex = self.normalize_color_to_hex(bg_color)

        command = [
            self.context.config.FFMPEG_PATH,
            '-y',
            '-f',
            'gif',
            '-trans_color', '0xff%s' % (bg_color_hex[1:] if bg_color_hex else 'ffffff'),
            '-i',
            input_file,
            '-filter_complex',
            'scale=trunc(iw/2)*2:trunc(ih/2)*2',
            '-an',
            '-movflags',
            'faststart',
            '-f',
            format,
            '-pix_fmt',
            'yuv420p'
        ]

        command += command_params

        command += [
            '-qmin',
            '10',
            '-qmax',
            '42',
            '-crf',
            '23',
            '-maxrate',
            '500k',
            output_file,
            '-loglevel',
            'error'
        ]

        with open(os.devnull) as null:
            subprocess.call(command, stdin=null)
        self.context.request.format = format

    def set_format(self):
        if 'webm' in self.context.request.filters:
            format = 'webm'
            command_params = [
                '-quality',
                'good',
                '-cpu-used',
                '4'
            ]
        else:
            format = 'mp4'
            command_params = [
                '-profile:v',
                'baseline',
                '-level',
                '4.0'
            ]
        return format, command_params

    def normalize_color_to_hex(self, color_string):
        try:
            return webcolors.normalize_hex("#" + color_string)
        except ValueError:
            pass

        try:
            return webcolors.name_to_hex(color_string)
        except ValueError:
            pass

        try:
            return webcolors.normalize_hex(color_string)
        except ValueError:
            pass

        if color_string:
            logger.error('background_color value could not be parsed')



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from thumbor.storages import BaseStorage
from tornado import gen


class Storage(BaseStorage):
    def __init__(self, context, file_storage=None, crypto_storage=None, detector_storage=None):
        BaseStorage.__init__(self, context)

        self.file_storage = file_storage

        self.crypto_storage = crypto_storage

        self.detector_storage = detector_storage

    def _init_file_storage(self):
        if self.file_storage is None:
            self.context.modules.importer.import_item(
                config_key='file_storage',
                item_value=self.context.config.MIXED_STORAGE_FILE_STORAGE,
                class_name='Storage'
            )
            self.file_storage = self.context.modules.file_storage = self.context.modules.importer.file_storage(self.context)

    def _init_crypto_storage(self):
        if self.crypto_storage is None:
            self.context.modules.importer.import_item(
                config_key='crypto_storage',
                item_value=self.context.config.MIXED_STORAGE_CRYPTO_STORAGE,
                class_name='Storage'
            )
            self.crypto_storage = self.context.modules.crypto_storage = self.context.modules.importer.crypto_storage(self.context)

    def _init_detector_storage(self):
        if self.detector_storage is None:
            self.context.modules.importer.import_item(
                config_key='detector_storage',
                item_value=self.context.config.MIXED_STORAGE_DETECTOR_STORAGE,
                class_name='Storage'
            )
            self.detector_storage = self.context.modules.detector_storage = \
                self.context.modules.importer.detector_storage(self.context)

    def put(self, path, bytes):
        self._init_file_storage()
        self.file_storage.put(path, bytes)

    def put_detector_data(self, path, data):
        self._init_detector_storage()
        self.detector_storage.put_detector_data(path, data)

    def put_crypto(self, path):
        self._init_crypto_storage()
        self.crypto_storage.put_crypto(path)

    @gen.coroutine
    def get_crypto(self, path):
        self._init_crypto_storage()
        result = yield gen.maybe_future(self.crypto_storage.get_crypto(path))
        raise gen.Return(result)

    @gen.coroutine
    def get_detector_data(self, path):
        self._init_detector_storage()
        result = yield gen.maybe_future(self.detector_storage.get_detector_data(path))
        raise gen.Return(result)

    @gen.coroutine
    def get(self, path):
        self._init_file_storage()
        result = yield gen.maybe_future(self.file_storage.get(path))
        raise gen.Return(result)

    @gen.coroutine
    def exists(self, path):
        self._init_file_storage()
        result = yield gen.maybe_future(self.file_storage.exists(path))
        raise gen.Return(result)

    def resolve_original_photo_path(self, request, filename):
        return self.file_storage.resolve_original_photo_path(request, filename)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from thumbor.filters import BaseFilter, filter_method
from thumbor.ext.filters import _colorize


class Filter(BaseFilter):

    @filter_method(BaseFilter.PositiveNumber, BaseFilter.PositiveNumber, BaseFilter.PositiveNumber, BaseFilter.String)
    def colorize(self, red_pct, green_pct, blue_pct, fill):
        fill_r, fill_g, fill_b = tuple(map(ord, fill.decode('hex')))
        mode, data = self.engine.image_data_as_rgb()
        imgdata = _colorize.apply(mode, red_pct, green_pct, blue_pct, fill_r, fill_g, fill_b, data)
        self.engine.set_image_data(imgdata)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import pkgutil
import pkg_resources

from thumbor import __version__

try:
    basestring        # Python 2
except NameError:
    basestring = str  # Python 3


class ErrorHandler(object):
    def __init__(self, config, client=None):
        import raven

        dsn = config.SENTRY_DSN_URL
        if not dsn:
            raise RuntimeError(
                "If you set USE_CUSTOM_ERROR_HANDLING to True, and you are using thumbor.error_handlers.sentry, " +
                "then you must specify the Sentry DSN using the SENTRY_DSN_URL configuration."
            )

        self.sentry = client or raven.Client(dsn)
        self.modules = self.get_modules()

    def get_modules(self):
        resolved = {}
        for _, module, _ in pkgutil.iter_modules():
            try:
                res_mod = pkg_resources.get_distribution(module)
                if res_mod is not None:
                    resolved[module] = res_mod.version
            except Exception:
                pass

        return resolved

    def handle_error(self, context, handler, exception):
        req = handler.request

        extra = {
            'thumbor-version': __version__
        }

        extra.update({
            'Headers': req.headers
        })

        cookies_header = extra.get('Headers', {}).get('Cookie', {})

        if isinstance(cookies_header, basestring):
            cookies = {}
            for cookie in cookies_header.split(';'):
                if not cookie:
                    continue
                values = cookie.strip().split('=')
                key, val = values[0], "".join(values[1:])
                cookies[key] = val
        else:
            cookies = cookies_header

        extra['Headers']['Cookie'] = cookies

        data = {
            'sentry.interfaces.Http': {
                'url': req.full_url(),
                'method': req.method,
                'data': req.arguments,
                'body': req.body,
                'query_string': req.query
            },
            'sentry.interfaces.User': {
                'ip': req.remote_ip,
            },
            'modules': self.modules
        }
        self.sentry.captureException(exception, extra=extra, data=data)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from os.path import expanduser, join
import tempfile

import derpconf.config as config
from derpconf.config import Config

from thumbor import __version__

try:
    basestring        # Python 2
except NameError:
    basestring = str  # Python 3

home = expanduser("~")

Config.define('THUMBOR_LOG_CONFIG', None, 'Logging configuration as json', 'Logging')
Config.define(
    'THUMBOR_LOG_FORMAT', '%(asctime)s %(name)s:%(levelname)s %(message)s',
    'Log Format to be used by thumbor when writing log messages.', 'Logging')

Config.define(
    'THUMBOR_LOG_DATE_FORMAT', '%Y-%m-%d %H:%M:%S',
    'Date Format to be used by thumbor when writing log messages.', 'Logging')

Config.define('MAX_WIDTH', 0, "Max width in pixels for images read or generated by thumbor", 'Imaging')
Config.define('MAX_HEIGHT', 0, "Max height in pixels for images read or generated by thumbor", 'Imaging')
Config.define('MAX_PIXELS', 75e6, "Max pixel count for images read by thumbor", "Imaging")
Config.define('MIN_WIDTH', 1, "Min width in pixels for images read or generated by thumbor", 'Imaging')
Config.define('MIN_HEIGHT', 1, "Min width in pixels for images read or generated by thumbor", 'Imaging')
Config.define('ALLOWED_SOURCES', [], "Allowed domains for the http loader to download. These are regular expressions.", 'Imaging')
Config.define('QUALITY', 80, 'Quality index used for generated JPEG images', 'Imaging')
Config.define('PROGRESSIVE_JPEG', True, 'Exports JPEG images with the `progressive` flag set.', 'Imaging')

Config.define('PILLOW_JPEG_SUBSAMPLING', None,
              'Specify subsampling behavior for Pillow (see `subsampling` \
              in http://pillow.readthedocs.org/en/latest/handbook/image-file-formats.html#jpeg).'
              'Be careful to use int for 0,1,2 and string for "4:4:4" notation. '
              'Will ignore `quality`. Using `keep` will copy the original file\'s subsampling.', 'Imaging')

Config.define('PILLOW_JPEG_QTABLES', None,
              'Specify quantization tables for Pillow (see `qtables` \
              in http://pillow.readthedocs.org/en/latest/handbook/image-file-formats.html#jpeg). '
              'Will ignore `quality`. Using `keep` will copy the original file\'s qtables.', 'Imaging')

Config.define('PILLOW_RESAMPLING_FILTER', 'LANCZOS',
              'Specify resampling filter for Pillow resize method.'
              'One of LANCZOS, NEAREST, BILINEAR, BICUBIC, HAMMING (Pillow>=3.4.0).', 'Imaging')

Config.define('WEBP_QUALITY', None, 'Quality index used for generated WebP images. If not set (None) the same level of '
              'JPEG quality will be used.', 'Imaging')

Config.define('PNG_COMPRESSION_LEVEL', 6, 'Compression level for generated PNG images.', 'Imaging')
Config.define('PILLOW_PRESERVE_INDEXED_MODE',
              True,
              'Indicates if final image should preserve indexed mode (P or 1) of original image', 'Imaging')
Config.define('AUTO_WEBP', False, 'Specifies whether WebP format should be used automatically if the request accepts it '
              '(via Accept header)', 'Imaging')
Config.define('AUTO_PNG_TO_JPG', False, 'Specifies whether a PNG image should be used automatically if the png image has '
              'no transparency (via alpha layer). '
              'WARNING: Depending on case, this is not a good deal. '
              'This transformation maybe causes distortions or the size of image can increase. '
              'Images with texts, for example, the result image maybe will be distorced. '
              'Dark images, for example, the size of result image maybe will be bigger. '
              'You have to evaluate the majority of your use cases to take a decision about the usage of this conf.', 'Imaging')
Config.define('SVG_DPI', 150,
              'Specify the ratio between 1in and 1px for SVG images. This is only used when'
              'rasterizing SVG images having their size units in cm or inches.', 'Imaging')
Config.define('MAX_AGE', 24 * 60 * 60, 'Max AGE sent as a header for the image served by thumbor in seconds', 'Imaging')
Config.define(
    'MAX_AGE_TEMP_IMAGE', 0,
    "Indicates the Max AGE header in seconds for temporary images (images with failed smart detection)", 'Imaging')
Config.define(
    'RESPECT_ORIENTATION', False,
    'Indicates whether thumbor should rotate images that have an Orientation EXIF header', 'Imaging')
Config.define(
    'IGNORE_SMART_ERRORS', False,
    'Ignore errors during smart detections and return image as a temp image (not saved in result storage and with '
    'MAX_AGE_TEMP_IMAGE age)', 'Imaging')

Config.define(
    'SEND_IF_MODIFIED_LAST_MODIFIED_HEADERS', False,
    'Sends If-Modified-Since & Last-Modified headers; requires support from result storage', 'Imaging')

Config.define(
    'PRESERVE_EXIF_INFO', False,
    'Preserves exif information in generated images. Increases image size in kbytes, use with caution.', 'Imaging')

Config.define(
    'ALLOW_ANIMATED_GIFS', True,
    'Indicates whether thumbor should enable the EXPERIMENTAL support for animated gifs.', 'Imaging')

Config.define(
    'USE_GIFSICLE_ENGINE', False,
    'Indicates whether thumbor should use gifsicle engine. Please note that smart cropping and filters are not '
    'supported for gifs using gifsicle (but won\'t give an error).', 'Imaging')

Config.define(
    'USE_BLACKLIST', False,
    'Indicates whether thumbor should enable blacklist functionality to prevent processing certain images.', 'Imaging')

Config.define(
    'ENGINE_THREADPOOL_SIZE', 0,
    'Size of the thread pool used for image transformations.  The default value is 0 (don\'t use a threadpoool. '
    'Increase this if you are seeing your IOLoop getting blocked (often indicated by your upstream HTTP '
    'requests timing out)', 'Imaging')

Config.define(
    'METRICS', 'thumbor.metrics.logger_metrics',
    'The metrics backend thumbor should use to measure internal actions. This must be the full name of a python module ' +
    '(python must be able to import it)', 'Extensibility')
Config.define(
    'LOADER', 'thumbor.loaders.http_loader',
    'The loader thumbor should use to load the original image. This must be the full name of a python module ' +
    '(python must be able to import it)', 'Extensibility')
Config.define(
    'STORAGE', 'thumbor.storages.file_storage',
    'The file storage thumbor should use to store original images. This must be the full name of a python module ' +
    '(python must be able to import it)', 'Extensibility')
Config.define(
    'RESULT_STORAGE', None,
    'The result storage thumbor should use to store generated images. This must be the full name of a python ' +
    'module (python must be able to import it)', 'Extensibility')
Config.define(
    'ENGINE', 'thumbor.engines.pil',
    'The imaging engine thumbor should use to perform image operations. This must be the full name of a ' +
    'python module (python must be able to import it)', 'Extensibility')

Config.define(
    'GIF_ENGINE', 'thumbor.engines.gif',
    'The gif engine thumbor should use to perform image operations. This must be the full name of a ' +
    'python module (python must be able to import it)', 'Extensibility')

Config.define('SECURITY_KEY', 'MY_SECURE_KEY', 'The security key thumbor uses to sign image URLs', 'Security')

Config.define('ALLOW_UNSAFE_URL', True, 'Indicates if the /unsafe URL should be available', 'Security')
Config.define('ALLOW_OLD_URLS', True, 'Indicates if encrypted (old style) URLs should be allowed', 'Security')
Config.define('ENABLE_ETAGS', True, 'Enables automatically generated etags', 'HTTP')
Config.define('MAX_ID_LENGTH', 32, 'Set maximum id length for images when stored', 'Storage')
Config.define('GC_INTERVAL', None, 'Set garbage collection interval in seconds', 'Performance')

# METRICS OPTIONS
Config.define('STATSD_HOST', None, 'Host to send statsd instrumentation to', 'Metrics')
Config.define('STATSD_PORT', 8125, 'Port to send statsd instrumentation to', 'Metrics')
Config.define('STATSD_PREFIX', None, 'Prefix for statsd', 'Metrics')

# FILE LOADER OPTIONS
Config.define('FILE_LOADER_ROOT_PATH', home, 'The root path where the File Loader will try to find images', 'File Loader')

# HTTP LOADER OPTIONS
Config.define(
    'HTTP_LOADER_CONNECT_TIMEOUT', 5,
    'The maximum number of seconds libcurl can take to connect to an image being loaded', 'HTTP Loader')
Config.define(
    'HTTP_LOADER_REQUEST_TIMEOUT', 20,
    'The maximum number of seconds libcurl can take to download an image', 'HTTP Loader')
Config.define(
    'HTTP_LOADER_FOLLOW_REDIRECTS', True,
    'Indicates whether libcurl should follow redirects when downloading an image', 'HTTP Loader')
Config.define(
    'HTTP_LOADER_MAX_REDIRECTS', 5,
    'Indicates the number of redirects libcurl should follow when downloading an image', 'HTTP Loader')
Config.define(
    'HTTP_LOADER_MAX_CLIENTS', 10,
    'The maximum number of simultaneous HTTP connections the loader can make before queuing', 'HTTP Loader')
Config.define(
    'HTTP_LOADER_FORWARD_USER_AGENT', False,
    'Indicates whether thumbor should forward the user agent of the requesting user', 'HTTP Loader')
Config.define(
    'HTTP_LOADER_FORWARD_ALL_HEADERS', False,
    'Indicates whether thumbor should forward the headers of the request', 'HTTP Loader')
Config.define(
    'HTTP_LOADER_FORWARD_HEADERS_WHITELIST', [],
    'Indicates which headers should be forwarded among all the headers of the request', 'HTTP Loader')
Config.define(
    'HTTP_LOADER_DEFAULT_USER_AGENT', "Thumbor/%s" % __version__,
    'Default user agent for thumbor http loader requests', 'HTTP Loader')
Config.define(
    'HTTP_LOADER_PROXY_HOST', None,
    'The proxy host needed to load images through', 'HTTP Loader')
Config.define(
    'HTTP_LOADER_PROXY_PORT', None,
    'The proxy port for the proxy host', 'HTTP Loader')
Config.define(
    'HTTP_LOADER_PROXY_USERNAME', None,
    'The proxy username for the proxy host', 'HTTP Loader')
Config.define(
    'HTTP_LOADER_PROXY_PASSWORD', None,
    'The proxy password for the proxy host', 'HTTP Loader')
Config.define(
    'HTTP_LOADER_CA_CERTS', None,
    'The filename of CA certificates in PEM format', 'HTTP Loader')
Config.define(
    'HTTP_LOADER_VALIDATE_CERTS', None,
    'Validate the server’s certificate for HTTPS requests', 'HTTP Loader')
Config.define(
    'HTTP_LOADER_CLIENT_KEY', None,
    'The filename for client SSL key', 'HTTP Loader')
Config.define(
    'HTTP_LOADER_CLIENT_CERT', None,
    'The filename for client SSL certificate', 'HTTP Loader')
Config.define(
    'HTTP_LOADER_CURL_ASYNC_HTTP_CLIENT', False,
    'If the CurlAsyncHTTPClient should be used', 'HTTP Loader')
Config.define(
    'HTTP_LOADER_CURL_LOW_SPEED_TIME', 0,
    'If HTTP_LOADER_CURL_LOW_SPEED_LIMIT and HTTP_LOADER_CURL_ASYNC_HTTP_CLIENT ' +
    'are set, then this is the time in seconds as integer after a download should ' +
    'timeout if the speed is below HTTP_LOADER_CURL_LOW_SPEED_LIMIT for that long')
Config.define(
    'HTTP_LOADER_CURL_LOW_SPEED_LIMIT', 0,
    'If HTTP_LOADER_CURL_LOW_SPEED_TIME and HTTP_LOADER_CURL_ASYNC_HTTP_CLIENT ' +
    'are set, then this is the limit in bytes per second as integer which should ' +
    'timeout if the speed is below that limit for HTTP_LOADER_CURL_LOW_SPEED_TIME seconds')

# FILE STORAGE GENERIC OPTIONS
Config.define(
    'STORAGE_EXPIRATION_SECONDS', 60 * 60 * 24 * 30,
    "Expiration in seconds for the images in the File Storage. Defaults to one month", 'File Storage')
Config.define(
    'STORES_CRYPTO_KEY_FOR_EACH_IMAGE', False,
    'Indicates whether thumbor should store the signing key for each image in the file storage. ' +
    'This allows the key to be changed and old images to still be properly found', 'File Storage')

# FILE STORAGE OPTIONS
Config.define(
    'FILE_STORAGE_ROOT_PATH', join(tempfile.gettempdir(), 'thumbor', 'storage'),
    'The root path where the File Storage will try to find images', 'File Storage')

# PHOTO UPLOAD OPTIONS
Config.define('UPLOAD_MAX_SIZE', 0, "Max size in Kb for images uploaded to thumbor", 'Upload')
Config.define('UPLOAD_ENABLED', False, 'Indicates whether thumbor should enable File uploads', 'Upload')
Config.define(
    'UPLOAD_PHOTO_STORAGE', 'thumbor.storages.file_storage',
    'The type of storage to store uploaded images with', 'Upload')
Config.define('UPLOAD_DELETE_ALLOWED', False, 'Indicates whether image deletion should be allowed', 'Upload')
Config.define('UPLOAD_PUT_ALLOWED', False, 'Indicates whether image overwrite should be allowed', 'Upload')
Config.define('UPLOAD_DEFAULT_FILENAME', 'image', 'Default filename for image uploaded', 'Upload')

# ALIASES FOR OLD PHOTO UPLOAD OPTIONS
Config.alias('MAX_SIZE', 'UPLOAD_MAX_SIZE')
Config.alias('ENABLE_ORIGINAL_PHOTO_UPLOAD', 'UPLOAD_ENABLED')
Config.alias('ORIGINAL_PHOTO_STORAGE', 'UPLOAD_PHOTO_STORAGE')
Config.alias('ALLOW_ORIGINAL_PHOTO_DELETION', 'UPLOAD_DELETE_ALLOWED')
Config.alias('ALLOW_ORIGINAL_PHOTO_PUTTING', 'UPLOAD_PUT_ALLOWED')

# MEMCACHE STORAGE OPTIONS
Config.define('MEMCACHE_STORAGE_SERVERS', ['localhost:11211'], 'List of Memcache storage server hosts', 'Memcache Storage')

# MIXED STORAGE OPTIONS
Config.define(
    'MIXED_STORAGE_FILE_STORAGE', 'thumbor.storages.no_storage',
    'Mixed Storage file storage. This must be the full name of a python module (python must be able ' +
    'to import it)', 'Mixed Storage')
Config.define(
    'MIXED_STORAGE_CRYPTO_STORAGE', 'thumbor.storages.no_storage',
    'Mixed Storage signing key storage. This must be the full name of a python module (python must be ' +
    'able to import it)', 'Mixed Storage')
Config.define(
    'MIXED_STORAGE_DETECTOR_STORAGE', 'thumbor.storages.no_storage',
    'Mixed Storage detector information storage. This must be the full name of a python module (python ' +
    'must be able to import it)', 'Mixed Storage')

# JSON META ENGINE OPTIONS
Config.define(
    'META_CALLBACK_NAME', None,
    'The callback function name that should be used by the META route for JSONP access', 'Meta')

# DETECTORS OPTIONS
Config.define(
    'DETECTORS', [],
    'List of detectors that thumbor should use to find faces and/or features. All of them must be ' +
    'full names of python modules (python must be able to import it)', 'Detection')

# FACE DETECTOR CASCADE FILE
Config.define(
    'FACE_DETECTOR_CASCADE_FILE', 'haarcascade_frontalface_alt.xml',
    'The cascade file that opencv will use to detect faces.', 'Detection')

Config.define(
    'GLASSES_DETECTOR_CASCADE_FILE', 'haarcascade_eye_tree_eyeglasses.xml',
    'The cascade file that opencv will use to detect glasses.', 'Detection')

Config.define(
    'PROFILE_DETECTOR_CASCADE_FILE', 'haarcascade_profileface.xml',
    'The cascade file that opencv will use to detect profile faces.', 'Detection')

Config.define(
    'OPTIMIZERS', [
        # 'thumbor.optimizers.jpegtran',
        # 'thumbor.optimizers.gifv',
    ], 'List of optimizers that thumbor will use to optimize images', 'Optimizers')

# OPTIMIZER CONFIGURATIONS
Config.define(
    'JPEGTRAN_PATH',
    '/usr/bin/jpegtran',
    'Path for the jpegtran binary',
    'Optimizers'
)

Config.define(
    'JPEGTRAN_SCANS_FILE',
    '',
    'Path for the progressive scans file to use with jpegtran optimizer. Implies progressive jpeg output',
    'Optimizers'
)

Config.define(
    'FFMPEG_PATH',
    '/usr/local/bin/ffmpeg',
    'Path for the ffmpeg binary used to generate gifv(h.264)',
    'Optimizers'
)

# AVAILABLE FILTERS
Config.define(
    'FILTERS', [
        'thumbor.filters.brightness',
        'thumbor.filters.colorize',
        'thumbor.filters.contrast',
        'thumbor.filters.rgb',
        'thumbor.filters.round_corner',
        'thumbor.filters.quality',
        'thumbor.filters.noise',
        'thumbor.filters.watermark',
        'thumbor.filters.equalize',
        'thumbor.filters.fill',
        'thumbor.filters.sharpen',
        'thumbor.filters.strip_exif',
        'thumbor.filters.strip_icc',
        'thumbor.filters.frame',
        'thumbor.filters.grayscale',
        'thumbor.filters.rotate',
        'thumbor.filters.format',
        'thumbor.filters.max_bytes',
        'thumbor.filters.convolution',
        'thumbor.filters.blur',
        'thumbor.filters.extract_focal',
        'thumbor.filters.focal',
        'thumbor.filters.no_upscale',
        'thumbor.filters.saturation',
        'thumbor.filters.max_age',
        'thumbor.filters.curve',
        'thumbor.filters.background_color',
        'thumbor.filters.upscale',
        'thumbor.filters.proportion',
        'thumbor.filters.stretch',
    ],
    'List of filters that thumbor will allow to be used in generated images. All of them must be ' +
    'full names of python modules (python must be able to import it)', 'Filters')

# RESULT STORAGE
Config.define(
    'RESULT_STORAGE_EXPIRATION_SECONDS', 0,
    'Expiration in seconds of generated images in the result storage', 'Result Storage')  # Never expires
Config.define(
    'RESULT_STORAGE_FILE_STORAGE_ROOT_PATH', join(tempfile.gettempdir(), 'thumbor', 'result_storage'),
    'Path where the Result storage will store generated images', 'Result Storage')
Config.define(
    'RESULT_STORAGE_STORES_UNSAFE', False,
    'Indicates whether unsafe requests should also be stored in the Result Storage', 'Result Storage')

# QUEUED DETECTOR REDIS OPTIONS
Config.define('REDIS_QUEUE_SERVER_HOST', 'localhost', 'Server host for the queued redis detector', 'Queued Redis Detector')
Config.define('REDIS_QUEUE_SERVER_PORT', 6379, 'Server port for the queued redis detector', 'Queued Redis Detector')
Config.define('REDIS_QUEUE_SERVER_DB', 0, 'Server database index for the queued redis detector', 'Queued Redis Detector')
Config.define('REDIS_QUEUE_SERVER_PASSWORD', None, 'Server password for the queued redis detector', 'Queued Redis Detector')

# QUEUED DETECTOR SQS OPTIONS
Config.define('SQS_QUEUE_KEY_ID', None, 'AWS key id', 'Queued SQS Detector')
Config.define('SQS_QUEUE_KEY_SECRET', None, 'AWS key secret', 'Queued SQS Detector')
Config.define('SQS_QUEUE_REGION', 'us-east-1', 'AWS SQS region', 'Queued SQS Detector')

# ERROR HANDLING
Config.define(
    'USE_CUSTOM_ERROR_HANDLING', False,
    'This configuration indicates whether thumbor should use a custom error handler.', 'Errors')
Config.define(
    'ERROR_HANDLER_MODULE', 'thumbor.error_handlers.sentry',
    'Error reporting module. Needs to contain a class called ErrorHandler with a ' +
    'handle_error(context, handler, exception) method.', 'Errors')

# SENTRY REPORTING MODULE
Config.define(
    'SENTRY_DSN_URL', '',
    'Sentry thumbor project dsn. i.e.: ' +
    'http://5a63d58ae7b94f1dab3dee740b301d6a:73eea45d3e8649239a973087e8f21f98@localhost:9000/2', 'Errors - Sentry')

# FILE REPORTING MODULE
Config.define('ERROR_FILE_LOGGER', None, 'File of error log as json', 'Errors')
Config.define('ERROR_FILE_NAME_USE_CONTEXT', False, 'File of error log name is parametrized with context attribute', 'Errors')

# SIGNER MODULE
Config.define(
    'URL_SIGNER', 'libthumbor.url_signers.base64_hmac_sha1',
    'The url signer thumbor should use to verify url signatures.' +
    'This must be the full name of a python module ' +
    '(python must be able to import it)', 'Extensibility'
)

# SERVER
Config.define(
    'MAX_WAIT_SECONDS_BEFORE_SERVER_SHUTDOWN', 0,
    'The amount of time to wait before shutting down the server, i.e. stop accepting requests.', 'Server'
)
Config.define(
    'MAX_WAIT_SECONDS_BEFORE_IO_SHUTDOWN', 0,
    'The amount of time to waut before shutting down all io, after the server has been stopped', 'Server'
)

Config.define(
    'APP_CLASS', 'thumbor.app.ThumborServiceApp',
    'Custom app class to override ThumborServiceApp. This config value is overridden by the -a command-line parameter.'
)


def generate_config():
    config.generate_config()


def format_value(value):
    if isinstance(value, basestring):
        return "'%s'" % value
    if isinstance(value, (tuple, list, set)):
        representation = '[\n'
        for item in value:
            representation += '#    %s' % item
        representation += '#]'
        return representation
    return value


if __name__ == '__main__':
    generate_config()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from thumbor.filters import BaseFilter, filter_method


class Filter(BaseFilter):

    @filter_method()
    def upscale(self):
        target_width = self.context.request.width
        target_height = self.context.request.height

        source_width, source_height = self.context.request.engine.size

        if source_width >= target_width or source_height >= target_height:
            return

        if source_width * 1.0 / target_width >= source_height * 1.0 / target_height:
            new_width = target_width
            new_height = int(round(source_height * target_width / source_width))
        else:
            new_width = int(round(source_width * target_height / source_height))
            new_height = target_height

        self.engine.resize(new_width, new_height)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki
# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
from thumbor.filters import BaseFilter, filter_method


class Filter(BaseFilter):
    @filter_method(BaseFilter.Boolean)
    def autojpg(self, enabled=True):
        self.context.request.auto_png_to_jpg = enabled



# -*- coding: utf-8 -*-

import base64
import hashlib
import hmac
from thumbor.url_signers import BaseUrlSigner

try:
    unicode        # Python 2
except NameError:
    unicode = str  # Python 3


class UrlSigner(BaseUrlSigner):
    """Validate urls and sign them using base64 hmac-sha1
    """

    def signature(self, url):
        return base64.urlsafe_b64encode(
            hmac.new(
                self.security_key, unicode(url).encode('utf-8'), hashlib.sha1
            ).digest()
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from subprocess import Popen, PIPE

from thumbor.optimizers import BaseOptimizer
from thumbor.utils import logger
from os.path import exists


class Optimizer(BaseOptimizer):

    def should_run(self, image_extension, buffer):
        if image_extension in ['.jpg', '.jpeg']:
            if not exists(self.context.config.JPEGTRAN_PATH):
                logger.warn(
                    'jpegtran optimizer enabled but binary JPEGTRAN_PATH does not exist')
                return False
            return True
        return False

    def run_optimizer(self, image_extension, buffer):
        if not self.should_run(image_extension, buffer):
            return buffer

        if 'strip_icc' in self.context.request.filters:
            copy_chunks = 'comments'
        else:
            # have to copy everything to preserve icc profile
            copy_chunks = 'all'

        command = [
            self.context.config.JPEGTRAN_PATH,
            '-copy',
            copy_chunks,
            '-optimize',
        ]

        if self.context.config.PROGRESSIVE_JPEG:
            command += [
                '-progressive'
            ]

        if self.context.config.JPEGTRAN_SCANS_FILE:
            if exists(self.context.config.JPEGTRAN_SCANS_FILE):
                command += [
                    '-scans',
                    self.context.config.JPEGTRAN_SCANS_FILE
                ]
            else:
                logger.warn('jpegtran optimizer scans file does not exist')

        # close_fds would have a sane default on Python 3 but with Python 2.7 it is False
        # per default but setting it to True + setting any of stdin, stdout, stderr will
        # make this crash on Windows
        # TODO remove close_fds=True if code was upgraded to Python 3
        jpg_process = Popen(command, stdin=PIPE, stdout=PIPE,
                            stderr=PIPE, close_fds=True)
        output_stdout, output_stderr = jpg_process.communicate(buffer)

        if jpg_process.returncode != 0:
            logger.warn('jpegtran finished with non-zero return code (%d): %s'
                        % (jpg_process.returncode, output_stderr))
            return buffer

        return output_stdout



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import os
from shutil import move
from json import dumps, loads
from datetime import datetime
from os.path import exists, dirname, getmtime, splitext
import hashlib
from uuid import uuid4

import thumbor.storages as storages
from thumbor.utils import logger
from tornado.concurrent import return_future


class Storage(storages.BaseStorage):

    def put(self, path, bytes):
        file_abspath = self.path_on_filesystem(path)
        temp_abspath = "%s.%s" % (file_abspath, str(uuid4()).replace('-', ''))
        file_dir_abspath = dirname(file_abspath)

        logger.debug('creating tempfile for %s in %s...' % (path, temp_abspath))

        self.ensure_dir(file_dir_abspath)

        with open(temp_abspath, 'w') as _file:
            _file.write(bytes)

        logger.debug('moving tempfile %s to %s...' % (temp_abspath, file_abspath))
        move(temp_abspath, file_abspath)

        return path

    def put_crypto(self, path):
        if not self.context.config.STORES_CRYPTO_KEY_FOR_EACH_IMAGE:
            return

        file_abspath = self.path_on_filesystem(path)
        file_dir_abspath = dirname(file_abspath)

        self.ensure_dir(file_dir_abspath)

        if not self.context.server.security_key:
            raise RuntimeError("STORES_CRYPTO_KEY_FOR_EACH_IMAGE can't be True if no SECURITY_KEY specified")

        crypto_path = '%s.txt' % splitext(file_abspath)[0]
        temp_abspath = "%s.%s" % (crypto_path, str(uuid4()).replace('-', ''))
        with open(temp_abspath, 'w') as _file:
            _file.write(self.context.server.security_key)

        move(temp_abspath, crypto_path)
        logger.debug('Stored crypto at %s (security key: %s)' % (crypto_path, self.context.server.security_key))

        return file_abspath

    def put_detector_data(self, path, data):
        file_abspath = self.path_on_filesystem(path)

        path = '%s.detectors.txt' % splitext(file_abspath)[0]
        temp_abspath = "%s.%s" % (path, str(uuid4()).replace('-', ''))

        file_dir_abspath = dirname(file_abspath)
        self.ensure_dir(file_dir_abspath)

        with open(temp_abspath, 'w') as _file:
            _file.write(dumps(data))

        move(temp_abspath, path)

        return file_abspath

    @return_future
    def get(self, path, callback):
        abs_path = self.path_on_filesystem(path)

        def file_exists(resource_available):
            if not resource_available:
                callback(None)
            else:
                with open(self.path_on_filesystem(path), 'r') as f:
                    callback(f.read())

        self.exists(None, file_exists, path_on_filesystem=abs_path)

    @return_future
    def get_crypto(self, path, callback):
        file_abspath = self.path_on_filesystem(path)
        crypto_file = "%s.txt" % (splitext(file_abspath)[0])

        if not exists(crypto_file):
            callback(None)
        else:
            with open(crypto_file, 'r') as crypto_f:
                callback(crypto_f.read())

    @return_future
    def get_detector_data(self, path, callback):
        file_abspath = self.path_on_filesystem(path)
        path = '%s.detectors.txt' % splitext(file_abspath)[0]

        def file_exists(resource_available):
            if not resource_available:
                callback(None)
            else:
                callback(loads(open(path, 'r').read()))
        self.exists(None, file_exists, path_on_filesystem=path)

    def path_on_filesystem(self, path):
        digest = hashlib.sha1(path.encode('utf-8')).hexdigest()
        return "%s/%s/%s" % (
            self.context.config.FILE_STORAGE_ROOT_PATH.rstrip('/'),
            digest[:2],
            digest[2:]
        )

    @return_future
    def exists(self, path, callback, path_on_filesystem=None):
        if path_on_filesystem is None:
            path_on_filesystem = self.path_on_filesystem(path)
        callback(os.path.exists(path_on_filesystem) and not self.__is_expired(path_on_filesystem))

    def remove(self, path):
        n_path = self.path_on_filesystem(path)
        return os.remove(n_path)

    def __is_expired(self, path):
        if self.context.config.STORAGE_EXPIRATION_SECONDS is None:
            return False
        timediff = datetime.now() - datetime.fromtimestamp(getmtime(path))
        return timediff.total_seconds() > self.context.config.STORAGE_EXPIRATION_SECONDS



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from os import fstat
from datetime import datetime
from os.path import join, exists, abspath

from six.moves.urllib.parse import unquote
from tornado.concurrent import return_future

from thumbor.loaders import LoaderResult


@return_future
def load(context, path, callback):
    file_path = join(
        context.config.FILE_LOADER_ROOT_PATH.rstrip('/'), path.lstrip('/'))
    file_path = abspath(file_path)
    inside_root_path = file_path.startswith(
        abspath(context.config.FILE_LOADER_ROOT_PATH))

    result = LoaderResult()

    if not inside_root_path:
        result.error = LoaderResult.ERROR_NOT_FOUND
        result.successful = False
        callback(result)
        return

    # keep backwards compatibility, try the actual path first
    # if not found, unquote it and try again
    if not exists(file_path):
        file_path = unquote(file_path)

    if exists(file_path):
        with open(file_path, 'r') as f:
            stats = fstat(f.fileno())

            result.successful = True
            result.buffer = f.read()

            result.metadata.update(
                size=stats.st_size,
                updated_at=datetime.utcfromtimestamp(stats.st_mtime))
    else:
        result.error = LoaderResult.ERROR_NOT_FOUND
        result.successful = False

    callback(result)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import base64
import hashlib

from Crypto.Cipher import AES

# Import the Signer class to support backward-compatibility
# and existing documentation
from thumbor.url_signers.base64_hmac_sha1 import UrlSigner as Signer  # NOQA
from thumbor.url import Url

try:
    unicode        # Python 2
except NameError:
    unicode = str  # Python 3

MODE_ECB = 1


class Cryptor(object):
    def __init__(self, security_key):
        self.security_key = (security_key * 16)[:16]

    def encrypt(self,
                width,
                height,
                smart,
                adaptive,
                full,
                fit_in,
                flip_horizontal,
                flip_vertical,
                halign,
                valign,
                trim,
                crop_left,
                crop_top,
                crop_right,
                crop_bottom,
                filters,
                image):

        generated_url = Url.generate_options(
            width=width,
            height=height,
            smart=smart,
            meta=False,
            adaptive=adaptive,
            full=full,
            fit_in=fit_in,
            horizontal_flip=flip_horizontal,
            vertical_flip=flip_vertical,
            halign=halign,
            valign=valign,
            trim=trim,
            crop_left=crop_left,
            crop_top=crop_top,
            crop_right=crop_right,
            crop_bottom=crop_bottom,
            filters=filters
        )

        url = "%s/%s" % (generated_url, hashlib.md5(image).hexdigest())

        def pad(s):
            return s + (16 - len(s) % 16) * "{"

        cipher = AES.new(self.security_key, MODE_ECB)
        encrypted = base64.urlsafe_b64encode(cipher.encrypt(pad(url.encode('utf-8'))))

        return encrypted

    def try_decrypt(self, encrypted_url_part, cryptor=None):
        decryptor = self
        if cryptor is not None:
            decryptor = cryptor

        try:
            opt = decryptor.decrypt(encrypted_url_part)
            if opt is not None:
                unicode(opt['image_hash'])
        except (UnicodeDecodeError, ValueError):
            opt = None

        return opt

    def get_options(self, encrypted_url_part, image_url):
        opt = self.try_decrypt(encrypted_url_part)

        # why this check for not self.security_key???
        # if opt is None and not self.security_key and self.context.config.STORES_CRYPTO_KEY_FOR_EACH_IMAGE:
        if opt is None and self.context.config.STORES_CRYPTO_KEY_FOR_EACH_IMAGE:
            security_key = self.context.modules.storage.get_crypto(image_url)

            if security_key is not None:
                cr = Cryptor(security_key)
                opt = self.try_decrypt(encrypted_url_part, cryptor=cr)

        if opt is None:
            return None

        image_hash = opt and opt.get('image_hash')
        image_hash = image_hash[1:] if image_hash and image_hash.startswith('/') else image_hash

        path_hash = hashlib.md5(image_url.encode('utf-8')).hexdigest()

        if not image_hash or image_hash != path_hash:
            return None

        opt['image'] = image_url
        opt['hash'] = opt['image_hash']
        del opt['image_hash']

        return opt

    def decrypt(self, encrypted):
        cipher = AES.new(self.security_key, MODE_ECB)

        try:
            debased = base64.urlsafe_b64decode(encrypted.encode("utf-8"))
            decrypted = cipher.decrypt(debased).rstrip('{')
        except TypeError:
            return None

        result = Url.parse_decrypted('/%s' % decrypted)

        result['image_hash'] = result['image']
        del result['image']

        return result



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from thumbor.ext.filters import _nine_patch
from thumbor.filters import BaseFilter, filter_method
from thumbor.loaders import LoaderResult
import tornado.gen


class Filter(BaseFilter):
    regex = r'(?:frame\((?P<url>.*?))'

    def on_image_ready(self, buffer):
        self.nine_patch_engine.load(buffer, None)
        self.nine_patch_engine.enable_alpha()
        self.engine.enable_alpha()

        nine_patch_mode, nine_patch_data = self.nine_patch_engine.image_data_as_rgb()
        padding = _nine_patch.get_padding(nine_patch_mode,
                                          nine_patch_data,
                                          self.nine_patch_engine.size[0],
                                          self.nine_patch_engine.size[1])

        self.handle_padding(padding)

        mode, data = self.engine.image_data_as_rgb()

        if mode != nine_patch_mode:
            raise RuntimeError('Image mode mismatch: %s != %s' % (
                mode, nine_patch_mode)
            )

        imgdata = _nine_patch.apply(mode,
                                    data,
                                    self.engine.size[0],
                                    self.engine.size[1],
                                    nine_patch_data,
                                    self.nine_patch_engine.size[0],
                                    self.nine_patch_engine.size[1])
        self.engine.set_image_data(imgdata)
        self.callback()

    def handle_padding(self, padding):
        '''Pads the image with transparent pixels if necessary.'''
        left = padding[0]
        top = padding[1]
        right = padding[2]
        bottom = padding[3]

        offset_x = 0
        offset_y = 0
        new_width = self.engine.size[0]
        new_height = self.engine.size[1]

        if left > 0:
            offset_x = left
            new_width += left
        if top > 0:
            offset_y = top
            new_height += top
        if right > 0:
            new_width += right
        if bottom > 0:
            new_height += bottom
        new_engine = self.context.modules.engine.__class__(self.context)
        new_engine.image = new_engine.gen_image((new_width, new_height), '#fff')
        new_engine.enable_alpha()
        new_engine.paste(self.engine, (offset_x, offset_y))
        self.engine.image = new_engine.image

    def on_fetch_done(self, result):
        # TODO if result.successful is False how can the error be handled?
        if isinstance(result, LoaderResult):
            buffer = result.buffer
        else:
            buffer = result

        self.nine_patch_engine.load(buffer, None)
        self.storage.put(self.url, self.nine_patch_engine.read())
        self.storage.put_crypto(self.url)
        self.on_image_ready(buffer)

    @filter_method(BaseFilter.String, async=True)
    @tornado.gen.coroutine
    def frame(self, callback, url):
        self.url = url
        self.callback = callback
        self.nine_patch_engine = self.context.modules.engine.__class__(self.context)
        self.storage = self.context.modules.storage

        buffer = yield tornado.gen.maybe_future(self.storage.get(self.url))
        if buffer is not None:
            self.on_image_ready(buffer)
        else:
            self.context.modules.loader.load(self.context, self.url, self.on_fetch_done)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import statsd
from thumbor.metrics import BaseMetrics


class Metrics(BaseMetrics):

    @classmethod
    def client(cls, config):
        """
        Cache statsd client so it doesn't do a DNS lookup
        over and over
        """
        if not hasattr(cls, "_client"):
            cls._client = statsd.StatsClient(config.STATSD_HOST, config.STATSD_PORT, config.STATSD_PREFIX)
        return cls._client

    def incr(self, metricname, value=1):
        Metrics.client(self.config).incr(metricname, value)

    def timing(self, metricname, value):
        Metrics.client(self.config).timing(metricname, value)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from os.path import join, dirname, abspath, isabs

import cv2
import numpy as np

from thumbor.point import FocalPoint
from thumbor.detectors import BaseDetector


class CascadeLoaderDetector(BaseDetector):

    def load_cascade_file(self, module_path, cascade_file_path):
        if not hasattr(self.__class__, 'cascade'):
            if isabs(cascade_file_path):
                cascade_file = cascade_file_path
            else:
                cascade_file = join(abspath(dirname(module_path)), cascade_file_path)
            self.__class__.cascade = cv2.CascadeClassifier(cascade_file)

    def get_min_size_for(self, size):
        ratio = int(min(size) / 15)
        ratio = max(20, ratio)
        return (ratio, ratio)

    def get_features(self):
        engine = self.context.modules.engine
        img = np.array(
            engine.convert_to_grayscale(
                update_image=False,
                with_alpha=False
            )
        )

        faces = self.__class__.cascade.detectMultiScale(
            img,
            1.2,
            4,
            minSize=self.get_min_size_for(engine.size)
        )
        faces_scaled = []

        for (x, y, w, h) in faces:
            faces_scaled.append((
                (
                    x.item(),
                    y.item(),
                    w.item(),
                    h.item()
                ), 0)
            )

        return faces_scaled

    def detect(self, callback):
        features = self.get_features()

        if features:
            for square, neighbors in features:
                self.context.request.focal_points.append(FocalPoint.from_square(*square))
            callback()
        else:
            self.next(callback)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from thumbor.ext.filters import _alpha
from thumbor.filters import BaseFilter, filter_method
from thumbor.loaders import LoaderResult
from thumbor.utils import logger
import tornado.gen
import math
import re


class Filter(BaseFilter):

    @staticmethod
    def detect_and_get_ratio_position(pos, length):
        match = re.match('^(-?)([0-9]+)p$', pos)

        if not match:
            return pos

        sign, ratio = match.groups()
        pos = "{sign}{pos}".format(sign=sign, pos=int(round(length * float(ratio) / 100, 0)))

        return pos

    @staticmethod
    def calc_watermark_size(sz, watermark_sz, w_ratio, h_ratio):
        wm_max_width = sz[0] * w_ratio if w_ratio else None
        wm_max_height = sz[1] * h_ratio if h_ratio else None

        if not wm_max_width:
            wm_max_width = watermark_sz[0] * wm_max_height / watermark_sz[1]

        if not wm_max_height:
            wm_max_height = watermark_sz[1] * wm_max_width / watermark_sz[0]

        if float(watermark_sz[0]) / wm_max_width >= float(watermark_sz[1]) / wm_max_height:
            wm_height = round(watermark_sz[1] * wm_max_width / watermark_sz[0])
            wm_width = round(wm_max_width)
        else:
            wm_height = round(wm_max_height)
            wm_width = round(watermark_sz[0] * wm_max_height / watermark_sz[1])

        return (wm_width, wm_height)

    def on_image_ready(self, buffer):
        self.watermark_engine.load(buffer, None)
        self.watermark_engine.enable_alpha()

        mode, data = self.watermark_engine.image_data_as_rgb()
        imgdata = _alpha.apply(mode,
                               self.alpha,
                               data)

        self.watermark_engine.set_image_data(imgdata)

        sz = self.engine.size
        watermark_sz = self.watermark_engine.size

        if self.w_ratio or self.h_ratio:
            watermark_sz = self.calc_watermark_size(sz, watermark_sz, self.w_ratio, self.h_ratio)
            self.watermark_engine.resize(watermark_sz[0], watermark_sz[1])

        self.x = self.detect_and_get_ratio_position(self.x, sz[0])
        self.y = self.detect_and_get_ratio_position(self.y, sz[1])

        mos_x = self.x == 'repeat'
        mos_y = self.y == 'repeat'
        center_x = self.x == 'center'
        center_y = self.y == 'center'
        if not center_x and not mos_x:
            inv_x = self.x[0] == '-'
            x = int(self.x)
        if not center_y and not mos_y:
            inv_y = self.y[0] == '-'
            y = int(self.y)

        if not mos_x:
            repeat_x = (1, 0)
            if center_x:
                x = (sz[0] - watermark_sz[0]) / 2
            elif inv_x:
                x = (sz[0] - watermark_sz[0]) + x
        else:
            repeat_x = divmod(sz[0], watermark_sz[0])
            if sz[0] * 1.0 / watermark_sz[0] < 2:
                repeat_x = (math.ceil(sz[0] * 1.0 / watermark_sz[0]), 10)
                space_x = 10
        if not mos_y:
            repeat_y = (1, 0)
            if center_y:
                y = (sz[1] - watermark_sz[1]) / 2
            elif inv_y:
                y = (sz[1] - watermark_sz[1]) + y
        else:
            repeat_y = divmod(sz[1], watermark_sz[1])
            if sz[1] * 1.0 / watermark_sz[1] < 2:
                repeat_y = (math.ceil(sz[1] * 1.0 / watermark_sz[1]), 10)
                space_y = 10

        if not mos_x and not mos_y:
            self.engine.paste(self.watermark_engine, (x, y), merge=True)
        elif mos_x and mos_y:
            if (repeat_x[0] * repeat_y[0]) > 100:
                tmpRepeatX = min(6, repeat_x[0])
                tmpRepeatY = min(6, repeat_y[0])
                repeat_x = (tmpRepeatX, sz[0] - tmpRepeatX * watermark_sz[0])
                repeat_y = (tmpRepeatY, sz[1] - tmpRepeatY * watermark_sz[1])
            space_x = repeat_x[1] / (max(repeat_x[0], 2) - 1)
            space_y = repeat_y[1] / (max(repeat_y[0], 2) - 1)
            for i in range(int(repeat_x[0])):
                x = i * space_x + i * watermark_sz[0]
                for j in range(int(repeat_y[0])):
                    y = j * space_y + j * watermark_sz[1]
                    self.engine.paste(self.watermark_engine, (x, y), merge=True)
        elif mos_x:
            space_x = repeat_x[1] / (max(repeat_x[0], 2) - 1)
            for i in range(int(repeat_x[0])):
                x = i * space_x + i * watermark_sz[0]
                self.engine.paste(self.watermark_engine, (x, y), merge=True)
        else:
            space_y = repeat_y[1] / (max(repeat_y[0], 2) - 1)
            for j in range(int(repeat_y[0])):
                y = j * space_y + j * watermark_sz[1]
                self.engine.paste(self.watermark_engine, (x, y), merge=True)

        self.callback()

    def on_fetch_done(self, result):
        if isinstance(result, LoaderResult) and not result.successful:
            logger.warn(
                'bad watermark result error=%s metadata=%s' %
                (result.error, result.metadata)
            )
            raise tornado.web.HTTPError(400)

        if isinstance(result, LoaderResult):
            buffer = result.buffer
        else:
            buffer = result

        self.storage.put(self.url, buffer)
        self.storage.put_crypto(self.url)
        self.on_image_ready(buffer)

    @tornado.gen.coroutine
    @filter_method(
        BaseFilter.String,
        r'(?:-?\d+p?)|center|repeat',
        r'(?:-?\d+p?)|center|repeat',
        BaseFilter.PositiveNumber,
        r'(?:-?\d+)|none',
        r'(?:-?\d+)|none',
        async=True
    )
    def watermark(self, callback, url, x, y, alpha, w_ratio=False, h_ratio=False):
        self.url = url
        self.x = x
        self.y = y
        self.alpha = alpha
        self.w_ratio = float(w_ratio) / 100.0 if w_ratio and w_ratio != 'none' else False
        self.h_ratio = float(h_ratio) / 100.0 if h_ratio and h_ratio != 'none' else False
        self.callback = callback
        self.watermark_engine = self.context.modules.engine.__class__(self.context)
        self.storage = self.context.modules.storage

        try:
            buffer = yield tornado.gen.maybe_future(self.storage.get(self.url))
            if buffer is not None:
                self.on_image_ready(buffer)
            else:
                self.context.modules.loader.load(self.context, self.url, self.on_fetch_done)
        except Exception as e:
            logger.exception(e)
            logger.warn("bad watermark")
            raise tornado.web.HTTPError(500)



# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import math

from thumbor.filters import BaseFilter, filter_method
from thumbor.ext.filters import _convolution

MAX_RADIUS = 150

try:
    xrange          # Python 2
except NameError:
    xrange = range  # Python 3


class Filter(BaseFilter):
    """
        Usage: /filters:blur(<radius> [, <sigma>])
        Examples of use:
            /filters:blur(1)/
            /filters:blur(4)/
            /filters:blur(4, 2)/
    """

    def generate_1d_matrix(self, sigma, radius):
        matrix_size = (radius * 2) + 1
        matrix = []
        two_sigma_squared = float(2 * sigma * sigma)
        for x in xrange(matrix_size):
            adj_x = x - radius
            exp = math.e ** -(((adj_x * adj_x)) / two_sigma_squared)
            matrix.append(exp / math.sqrt(two_sigma_squared * math.pi))
        return tuple(matrix), matrix_size

    @filter_method(BaseFilter.PositiveNonZeroNumber, BaseFilter.DecimalNumber)
    def blur(self, radius, sigma=0):
        if sigma == 0:
            sigma = radius
        if radius > MAX_RADIUS:
            radius = MAX_RADIUS
        matrix, matrix_size = self.generate_1d_matrix(sigma, radius)
        mode, data = self.engine.image_data_as_rgb()
        imgdata = _convolution.apply(mode, data, self.engine.size[0], self.engine.size[1], matrix, matrix_size, True)
        imgdata = _convolution.apply(mode, imgdata, self.engine.size[0], self.engine.size[1], matrix, 1, True)
        self.engine.set_image_data(imgdata)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from thumbor.filters import BaseFilter, filter_method


class Filter(BaseFilter):

    @filter_method(r'[\w]+')
    def background_color(self, color):
        self.fill_engine = self.engine.__class__(self.context)
        x, y = self.engine.size
        try:
            self.fill_engine.image = self.fill_engine.gen_image((x, y), color)
        except (ValueError, RuntimeError):
            self.fill_engine.image = self.fill_engine.gen_image((x, y), '#%s' % color)
        self.fill_engine.paste(self.engine, (0, 0), merge=True)
        self.engine.image = self.fill_engine.image



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com


from thumbor.detectors.queued_detector import QueuedDetector


class Detector(QueuedDetector):
    detection_type = 'face'



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import os
from os.path import exists
from tornado.concurrent import return_future

from thumbor.loaders import LoaderResult
from thumbor.engines import BaseEngine


class ResultStorageResult(LoaderResult):
    @property
    def last_modified(self):
        '''
        Retrieves last_updated metadata if available
        :return:
        '''
        return self.metadata.get('LastModified', None)

    @property
    def mime(self):
        '''
        Retrieves mime metadata if available
        :return:
        '''
        return self.metadata['ContentType'] if 'ContentType' in self.metadata else BaseEngine.get_mimetype(self.buffer)

    def __len__(self):
        return self.metadata['ContentLength'] if 'ContentLength' in self.metadata else len(self.buffer)


class BaseStorage(object):
    def __init__(self, context):
        self.context = context

    def put(self, bytes):
        raise NotImplementedError()

    @return_future
    def get(self, callback):
        raise NotImplementedError()

    def last_updated(self):
        raise NotImplementedError()

    def ensure_dir(self, path):
        if not exists(path):
            try:
                os.makedirs(path)
            except OSError as err:
                # FILE ALREADY EXISTS = 17
                if err.errno != 17:
                    raise



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com timehome@corp.globo.com

try:
    unicode        # Python 2
except NameError:
    unicode = str  # Python 3


class BaseUrlSigner(object):
    def __init__(self, security_key):
        if isinstance(security_key, unicode):
            security_key = security_key.encode('utf-8')
        self.security_key = security_key

    def validate(self, actual_signature, url):
        url_signature = self.signature(url)
        return url_signature == actual_signature

    def signature(self, url):
        raise NotImplementedError()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from os.path import abspath, dirname, join

import cv2
import numpy as np

from thumbor.filters import BaseFilter, filter_method

CASCADE_FILE_PATH = abspath(join(dirname(__file__), 'haarcascade_eye.xml'))

MIN_SIZE = (20, 20)
HAAR_SCALE = 1.2
MIN_NEIGHBORS = 3
RED_THRESHOLD = 2.0


class Filter(BaseFilter):
    @filter_method()
    def red_eye(self):
        faces = [face for face in self.context.request.focal_points if face.origin == 'Face Detection']
        if not faces:
            return

        mode, data = self.engine.image_data_as_rgb()
        mode = mode.lower()
        size = self.engine.size

        image = np.ndarray(shape=(size[1], size[0], 4 if mode == 'rgba' else 3), dtype='|u1', buffer=data)
        image.setflags(write=1)

        for face in faces:
            face_x = int(face.x - face.width / 2)
            face_y = int(face.y - face.height / 2)

            face_image = image[face_y:face_y + face.height, face_x:face_x + face.width]

            eyeRects = self.cascade.detectMultiScale(
                face_image,
                scaleFactor=HAAR_SCALE,
                minNeighbors=MIN_NEIGHBORS,
                minSize=MIN_SIZE)

            for x, y, w, h in eyeRects:
                # Crop the eye region
                eyeImage = face_image[y:y + h, x:x + w]

                # split the images into 3 channels
                r, g, b = cv2.split(eyeImage)

                # Add blue and green channels
                bg = cv2.add(b, g)
                mean = bg / 2
                # threshold the mask based on red color and combination of blue and green color
                mask = ((r > RED_THRESHOLD * mean) & (r > 60)).astype(np.uint8) * 255
                # Some extra region may also get detected , we find the largest region
                # find all contours
                contours_return = cv2.findContours(mask.copy(), cv2.RETR_EXTERNAL,
                                                   cv2.CHAIN_APPROX_NONE)  # It return contours and Hierarchy

                if len(contours_return) == 2:
                    contours, _ = contours_return
                else:
                    _, contours, _ = contours_return

                # find contour with max area
                maxArea = 0
                maxCont = None
                for cont in contours:
                    area = cv2.contourArea(cont)
                    if area > maxArea:
                        maxArea = area
                        maxCont = cont

                if maxCont is None:
                    continue

                mask = mask * 0  # Reset the mask image to complete black image
                # draw the biggest contour on mask
                cv2.drawContours(mask, [maxCont], 0, (255), -1)
                # Close the holes to make a smooth region
                mask = cv2.morphologyEx(mask, cv2.MORPH_CLOSE, cv2.getStructuringElement(cv2.MORPH_DILATE, (5, 5)))
                mask = cv2.dilate(mask, (3, 3), iterations=3)

                # The information of only red color is lost,
                # So we fill the mean of blue and green color in all three channels(BGR) to maintain the texture
                # Fill this black mean value to masked image
                mean = cv2.bitwise_and(mean, mask)  # mask the mean image
                mean = cv2.cvtColor(mean, cv2.COLOR_GRAY2RGB)  # convert mean to 3 channel
                mask = cv2.cvtColor(mask, cv2.COLOR_GRAY2RGB)  # convert mask to 3 channel
                eye = cv2.bitwise_and(~mask, eyeImage) + mean  # Copy the mean color to masked region to color image
                face_image[y:y + h, x:x + w] = eye

        self.engine.set_image_data(image.tobytes())

    @property
    def cascade(self):
        if not hasattr(self, '_cascade'):
            setattr(self, '_cascade', cv2.CascadeClassifier(CASCADE_FILE_PATH))

        return getattr(self, '_cascade')



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import cv2
import numpy as np

from thumbor.detectors import BaseDetector
from thumbor.point import FocalPoint
from thumbor.utils import logger


class Detector(BaseDetector):

    def detect(self, callback):
        engine = self.context.modules.engine
        try:
            img = np.array(
                engine.convert_to_grayscale(
                    update_image=False,
                    with_alpha=False
                )
            )
        except Exception as e:
            logger.exception(e)
            logger.warn('Error during feature detection; skipping to next detector')
            self.next(callback)
            return

        points = cv2.goodFeaturesToTrack(
            img,
            maxCorners=20,
            qualityLevel=0.04,
            minDistance=1.0,
            useHarrisDetector=False,
        )
        if points is not None:
            for point in points:
                x, y = point.ravel()
                self.context.request.focal_points.append(FocalPoint(x.item(), y.item(), 1))
            callback()
        else:
            self.next(callback)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com
import datetime

from thumbor.handlers import ImageApiHandler
from thumbor.engines import BaseEngine
import tornado.gen as gen
import tornado.web


##
# Handler to retrieve or modify existing images
# This handler support GET, PUT and DELETE method to manipulate existing images
##
class ImageResourceHandler(ImageApiHandler):

    @gen.coroutine
    def check_resource(self, id):
        id = id[:self.context.config.MAX_ID_LENGTH]
        # Check if image exists
        exists = yield gen.maybe_future(self.context.modules.storage.exists(id))

        if exists:
            body = yield gen.maybe_future(self.context.modules.storage.get(id))
            self.set_status(200)

            mime = BaseEngine.get_mimetype(body)
            if mime:
                self.set_header('Content-Type', mime)

            max_age = self.context.config.MAX_AGE
            if max_age:
                self.set_header('Cache-Control', 'max-age=' + str(max_age) + ',public')
                self.set_header('Expires', datetime.datetime.utcnow() + datetime.timedelta(seconds=max_age))
            self.write(body)
            self.finish()
        else:
            self._error(404, 'Image not found at the given URL')

    def put(self, id):
        id = id[:self.context.config.MAX_ID_LENGTH]
        # Check if image overwriting is allowed
        if not self.context.config.UPLOAD_PUT_ALLOWED:
            self._error(405, 'Unable to modify an uploaded image')
            return

        # Check if the image uploaded is valid
        if self.validate(self.request.body):
            self.write_file(id, self.request.body)
            self.set_status(204)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def delete(self, id):
        id = id[:self.context.config.MAX_ID_LENGTH]
        # Check if image deleting is allowed
        if not self.context.config.UPLOAD_DELETE_ALLOWED:
            self._error(405, 'Unable to delete an uploaded image')
            return

        # Check if image exists
        exists = yield gen.maybe_future(self.context.modules.storage.exists(id))
        if exists:
            self.context.modules.storage.remove(id)
            self.set_status(204)
        else:
            self._error(404, 'Image not found at the given URL')

    @tornado.web.asynchronous
    def get(self, id):
        self.check_resource(id)

    @tornado.web.asynchronous
    def head(self, id):
        self.check_resource(id)
